#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

void handler(int signo, siginfo_t *info, void *v)
{
    /* nothing to do, actually */
}

#define BSIZE 1024

int main(int argc, char **argv)
{
    char buf[BSIZE];
    char *s, dirname[PATH_MAX] = "./";
    int fd, dd, i;
    struct sigaction sa;
    struct stat st;

    if (argc != 2) {
	fprintf(stderr, "%s: Use: \"%s <file>\"\n", argv[0], argv[0]);
	exit(1);
    }

    /* check that the file exists and is regular, then open it */
    if (stat(argv[1], &st)) {
	fprintf(stderr, "%s: %s: %s\n", argv[0], argv[1], strerror(errno));
	exit(1);
    }
    if (!S_ISREG(st.st_mode)) {
	fprintf(stderr, "%s: %s: not a regular file\n", argv[0], argv[1]);
	exit(1);
    }
    if ((fd = open(argv[1], O_RDONLY)) < 0) {
	fprintf(stderr, "%s: %s: %s\n", argv[0], argv[1], strerror(errno));
	exit(1);
    }

    /* build the directory name and open it*/
    if (argv[1][0] == '/')
	strncpy(dirname, argv[1], PATH_MAX-1);
    else
	strncat(dirname, argv[1], PATH_MAX-3);
    s = strrchr(dirname, '/');
    if (s == dirname) s++; /* a file in root dir */
    *s = '\0';
    if ((dd = open(dirname, O_RDONLY)) < 0) {
	fprintf(stderr, "%s: %s: %s\n", argv[0], dirname, strerror(errno));
	exit(1);
    }

    /* prepare signal handler for SIGUSR1 */
    memset(&sa, 0, sizeof(sa));
    sa.sa_sigaction = handler;
    sa.sa_flags = SA_SIGINFO;
    if (sigaction(SIGUSR1, &sa, NULL)) {
	fprintf(stderr, "%s: sigaction(): %s\n", argv[0], strerror(errno));
	exit(1);
    }
	
    /* elect SIGUSR1 for notification, and enable notification */
    if (fcntl(dd, F_SETSIG, SIGUSR1)) {
	fprintf(stderr, "%s: %s: F_SETSIG: %s\n", argv[0], dirname,
	       strerror(errno));
	exit(1);
    }
    if (fcntl(dd, F_NOTIFY, DN_MODIFY | DN_MULTISHOT)) {
	fprintf(stderr, "%s: %s: F_NOTIFY: %s\n", argv[0], dirname,
	       strerror(errno));
	exit(1);
    }

    /* seek to end of file and loop waiting for EINTR, then read */
    lseek(fd, 0, SEEK_END);
    while (1) {
	select(0, NULL, NULL, NULL, NULL); /* forever */
	if (errno != EINTR) abort(); /* can't happen */
	while ((i = read(fd, buf, BSIZE)) > 0) {
	    write(fileno(stdout), buf, i);
	}
    }
    exit(0);
}
